from ..utils.filesystem_utils import get_file_contents
from ..utils.configuration import Configuration

class Report:
    class Status:
        FAILURE = "FAILURE"
        SUCCESS = "SUCCESS"

    REPORT_LOCATION = Configuration.get_configuration()[Configuration.Key.OUTPUT_REPORT_LOCATION]
    TEMPLATE_LOCATION = Configuration.get_configuration()[Configuration.Key.REPORT_TEMPLATE_LOCATION]

    @staticmethod
    def __get_report_template():
        return get_file_contents(Report.TEMPLATE_LOCATION)

    @staticmethod
    def __save_report(report, path):
        with open(path, 'w+') as f:
            f.write(report)

    @staticmethod
    def write_report(status: Status):
        template = Report.__get_report_template()
        report = template.format(status=status)
        Report.__save_report(report, Report.REPORT_LOCATION)



import os, shutil, time

dist_dir = "./dist/"

if not os.path.exists("./src"):
    raise Exception("You should run this script in the python-ep-algorithm-wrapper folder!")

if not os.path.exists("./.dev/local.creds.txt"):
    raise Exception("No credentials file")

def clear_folder(folder):
    for filename in os.listdir(folder):
        file_path = os.path.join(folder, filename)
        try:
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)
        except Exception as e:
            print('Failed to delete %s. Reason: %s' % (file_path, e))

if not os.path.exists(dist_dir):
    os.makedirs(dist_dir)
else:
    clear_folder(dist_dir)

with open("./.dev/local.creds.txt", 'r') as f:
    creds = f.read().splitlines()

username = creds[0]
password = creds[1]

os.system("python -m pip uninstall -y epalgorithmwrapper")

os.system("python -m build")

os.system(f"python -m twine upload -u {username} -p {password} dist/*")

print("Manually reinstall eputils!")


from ..utils.context import Context
from .message_processor import MessageProcessor
from .instruction import JobInstruction
from .job_provider import JobProvider

class StaticJobProvider(JobProvider):
    def __init__(self, 
            context: Context, 
            message_processor: MessageProcessor, 
            reference,
            patient_id:str,
            case_id:str):
        super().__init__()

        self._context = context
        self._message_processor = message_processor
        self._reference = reference
        self._patient_id = patient_id
        self._case_id = case_id

        self.init_job()

    def init_job(self):
        instruction = JobInstruction(self._reference, self._patient_id, self._case_id)
        self._message_processor.process(instruction)
    


from enum import Enum
from .filesystem_utils import get_file_contents
import json

class Manifest(dict):
    MANIFEST_FILE_PATH = "./manifest.json"

    class Key(Enum):
        MANIFEST_VERSION = "manifest-version"
        IDENTIFIER = "identifier"
        COMMAND = "command"
        DEVELOPMENT = "dev"
        INPUT = "input"
        OUTPUT = "output"
    
    def __init__(self, dict: dict) -> None:
        self.__dict__ = dict

    def __setitem__(self, key, item):
        self.__dict__[key.value] = item

    def __getitem__(self, key):
        return self.__dict__[key.value]

    def __repr__(self):
        return repr(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def __delitem__(self, key):
        del self.__dict__[key.value]
    
    def __contains__(self, key) -> bool:
        return key.value in self.__dict__

    def clear(self):
        return self.__dict__.clear()

    def copy(self):
        return self.__dict__.copy()

    def has_key(self, k):
        return k.value in self.__dict__

    def update(self, *args, **kwargs):
        return self.__dict__.update(*args, **kwargs)

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()   
             
    @staticmethod
    def get_manifest(development=False):
        configuration_str = get_file_contents(Manifest.MANIFEST_FILE_PATH)
        mf: dict = json.loads(configuration_str)
        if "dev" in mf:
            if development:
                dev_vars: dict = mf["dev"]
                for k, v in dev_vars.items():
                    mf[k] = v
            mf.pop("dev")

        return Manifest(mf)


import requests

TIMEOUT=2

def _log(logger, msg):
    if logger is None: return
    logger.info(msg)

def send_post(url, json_body, logger=None):
    s = requests.Session()
    s.trust_env=False

    _log(logger, "Sending post")
    try:
        r = s.post(url, json=json_body, verify=False, timeout=TIMEOUT)
    except Exception:
        return None

    return r

def send_post_raw(url, body, logger=None, custom_headers = {}):
    s = requests.Session()
    s.trust_env=False

    _log(logger, "Sending post")
    try:
        r = s.post(url, data=body, verify=False, headers=custom_headers, timeout=TIMEOUT)
    except Exception:
        return None

    return r

def send_post_file(url, file, logger=None):
    s = requests.Session()
    s.trust_env=False

    files = {'file': open(file, 'rb')}
    _log(logger, "Sending post file")

    r = s.post(url, files=files)

    return r


import logging
from ..utils.http_tools import send_post

class DataProvider:
    def send_request(self, request_body):
        """Send a request with a certain body"""
    
class HTTPDataProvider(DataProvider):
    def __init__(self, download_url: str, logger: logging.Logger) -> None:
        self.download_url = download_url
        self.logger = logger
    
    def send_request(self, request_body):
        self.logger.info("Sending request..")
        r = send_post(self.download_url, request_body, self.logger)

        if r is None or not r.ok:
            self.logger.error(f"!Could not download resource: {request_body}")
            return None
        
        self.logger.info("Received data")
        return r.content

class LocalFileProvider(DataProvider):
    def __init__(self, logger: logging.Logger) -> None:
        self.logger = logger

    def send_request(self, request_body):
        self.logger.info("Using local file instead of downloading")
        return True # File is already available



from .dataprovider import DataProvider, HTTPDataProvider, LocalFileProvider
import logging

def get_dataprovider(host: str, logger: logging.Logger, local=False) -> DataProvider:
    if local:
        logger.warning("Using local data provider instead of HTTP provider")
        return __get_local_dataprovider(logger)
    logger.info("Using normal data provider")
    return __get_http_dataprovider(host, logger)

def __get_http_dataprovider(host: str, logger: logging.Logger):
    url = f"http://{host}"
    download_url = f"{url}/download"
    return HTTPDataProvider(download_url, logger)

def __get_local_dataprovider(logger: logging.Logger):
    return LocalFileProvider(logger)


from ..utils.manifest import Manifest
from ..utils.configuration import Configuration
from ..utils.cli_tools import dispatch_command
from ..utils.context import Context
from ..utils.filesystem_utils import clear_folder, get_file_contents
from ..utils.message import get_message_formatter
from ..reporter.reporter_factory import get_reporter
from .instruction import JobInstruction
from .resource_manager import ResourceManager
import os

class MessageProcessor:
    def process(self, instruction: JobInstruction):
        """Process message with self-defined logic"""

class JsonV1Processor(MessageProcessor):
    def __init__(self, context: Context, resource_manager: ResourceManager) -> None:
        self.resource_manager = resource_manager
        self._logger = context.logger
        self._context = context
        self._message_formatter = get_message_formatter()
        self._reporter = get_reporter(context.connection.host_algorithm_manager, context.logger, context.input_provider)

    def _print(self, message):
        self._logger.info(message)

    def process(self, instruction: JobInstruction):
        manifest = self._context.manifest
        configuration = self._context.configuration

        reference = instruction.reference
        input_resources = manifest[Manifest.Key.INPUT]
        output_resources = manifest[Manifest.Key.OUTPUT]

        self._print(f"Starting with job: {reference}")
        self._print(f"With input: {input_resources}")
        self._print(f"Will output: {output_resources}")

        self._print("Putting sub-job-reference in environment variables..")
        os.environ["EP_ALGORITHM_SUB_JOB_REFERENCE"] = reference

        self._print("Clearing previous output..")
        clear_folder(self._context.configuration[Configuration.Key.OUTPUT_DIRECTORY])
        self._print("Cleared!")

        input_provider = self._context.input_provider

        self._print("Fetching all resources..")
        for resource in input_resources:

            resource_id = resource["name"]
            input_resources_context: dict = input_provider.get_input_resources_context()
            if resource_id not in input_resources_context:
                raise Exception(f"Expected {resource_id} in input resource context, but only has {input_resources_context.keys()}")
            
            resource_spec = input_resources_context[resource_id]
            resource_spec_patientcontext = resource_spec["patient-context"]
            resource_spec_patient_id = resource_spec_patientcontext["patient-id"]
            resource_spec_case_id = resource_spec_patientcontext["case-id"]
            resource_spec_resourcecontext = resource_spec["resource-context"]

            resource_instruction = {
                "store": resource["store"],
                "resource-identification": resource_id,
                "resource-filename": resource["data"],
                "resource-context": resource_spec_resourcecontext,
                "patient-context": {
                    "patient-id": resource_spec_patient_id,
                    "case-id": resource_spec_case_id
                }
            }

            self._print(f"Fetching resource {resource_instruction}..")

            fetched = self.resource_manager.fetch_resource(resource_instruction, self._context.configuration[Configuration.Key.INPUT_DIRECTORY])
            if not fetched:
                self._print(f"Could not fetch {resource}")
                self._publish_message(
                    self._message_formatter.get_message(reference, "failed", f"Could not download resource: {resource}"))
                return
            self._print("Fetched!")
        self._print("Fetched all!")

        self._print("Reading algorithm manifest..")
        command = manifest[Manifest.Key.COMMAND]
        
        self._print(f"Dispatching algorithm with command `{command}`..\n---------------\n")
        dispatch_command(command)
        self._print(f"\n---------------\nAlgorithm finished!")

        # TODO: Read report
        self._print("Reading algorithm execution report..")
        report = get_file_contents(configuration[Configuration.Key.OUTPUT_REPORT_LOCATION])
        self._print(f"Todo: process {report}")

        if report is None:
            status = "failed"
            description = "Algorithm failed!"
        else:
            status = "busy"
            description = "Algorithm finished! Writing back output.."
        
        self._print(description)

        # TODO: Send back data
        self._publish_message(
            self._message_formatter.get_message(reference, status, description))
        
        if status == "failed": return

        patient_id = input_provider.get_patient_id()
        case_id = input_provider.get_case_id()
        resource_context = f"job:{input_provider.get_reference()}"

        for resource in output_resources:
            resource_instruction = {
                "store": resource["store"],
                "resource-identification": resource["name"],
                "resource-filename": resource["data"],
                "resource-context": resource_context,
                "patient-context": {
                    "patient-id": patient_id,
                    "case-id": case_id
                }
            }
            self._print(f"Uploading resource {resource_instruction}..")

            posted = self.resource_manager.post_resource(resource_instruction, self._context.configuration[Configuration.Key.OUTPUT_DIRECTORY])

            if not posted:
                self._print("Could not post!! Failing")
                self._publish_message(
                    self._message_formatter.get_message(reference, "failed", "Could not upload result"))
                return

            self._print("Uploaded!")

        self._publish_message(
            self._message_formatter.get_message(reference, "finished", "Process completed"))
    
    def _publish_message(self, message: str):
        self._reporter.report(message)
        


import os
from queue import Queue
import logging

from .utils.manifest import Manifest
from .utils.configuration import Configuration
from .utils.context import Context, ConnectionContext, MessagebusContext
from .jobprovider.job_provider_factory import get_job_provider
from .jobprovider.message_processor import JsonV1Processor
from .jobprovider.resource_manager import get_resource_manager
from .connection import get_messagebroker_host, get_filesystem_host, get_creds_provider, get_algorithm_manager_host
from .utils.input_provider import InputProvider
from .dataprovider.dataprovider_factory import get_dataprovider
from .datauploader.datauploader_factory import get_datauploader

context = None
job_provider = None
publish_queue = None

def decorated_print(header, message):
    return f"{header}: {message}"

def start(logger: logging.Logger, input_provider: InputProvider):
    is_development = input_provider.is_development_mode()
    local_data_mode = input_provider.skip_download_upload()
    logger.warning(f"Development mode: {is_development}")

    host_messagebroker = get_messagebroker_host()
    host_filesystem = get_filesystem_host()
    host_algorithm_manager = get_algorithm_manager_host()

    logger.info(f"Using host messagebroker {host_messagebroker}")
    logger.info(f"Using host filesystem {host_filesystem}")

    data_provider = get_dataprovider(host_filesystem, logger, local_data_mode)
    data_uploader = get_datauploader(host_filesystem, logger, local_data_mode)
    
    creds_provider = get_creds_provider()
    configuration = Configuration.get_configuration()
    manifest = Manifest.get_manifest(is_development)

    connection = ConnectionContext(host_messagebroker, 
                                   host_filesystem,
                                   host_algorithm_manager,
                                   creds_provider)
    messagebus = MessagebusContext(app_id=f"algorithm.{manifest[Manifest.Key.IDENTIFIER]}", 
                                   job_queue_name=manifest[Manifest.Key.IDENTIFIER],
                                   status_update_exchange=configuration[Configuration.Key.RABBITMQ_STATUS_UPDATE_EXCHANGE], 
                                   status_update_routing=configuration[Configuration.Key.RABBITMQ_STATUS_UPDATE_ROUTING])
    context = Context(connection,
                      messagebus,
                      logger, 
                      manifest,
                      configuration,
                      input_provider,
                      data_provider,
                      data_uploader)

    logger.info(decorated_print("Manifest", manifest))
    logger.info(decorated_print("Configuration", configuration))

    resource_manager = get_resource_manager(context)
    message_processor = JsonV1Processor(context, resource_manager)
    get_job_provider(input_provider, context, message_processor)

def get_publish_queue() -> Queue:
    return publish_queue

def get_context() -> Context:
    return context


from .rest_api_reporter import RESTAPIReporter, PreprocessRESTAPIReporter, VoidReporter
from .reporter import Reporter
from ..utils.input_provider import InputProvider

def get_reporter(host, logger, input_provider: InputProvider) -> Reporter:
    if input_provider is not None and input_provider.skip_reports():
        return VoidReporter(logger)
    return RESTAPIReporter(host, logger)

def get_preprocess_reporter(host, logger, input_provider: InputProvider) -> Reporter:
    if input_provider is not None and input_provider.skip_reports():
        return VoidReporter(logger)
    return PreprocessRESTAPIReporter(host, logger)


from .datauploader import DataUploader, HTTPDataUploader, VoidUploader
import logging

def get_datauploader(host: str, logger: logging.Logger, local=False) -> DataUploader:
    upload_url = f"http://{host}"
    if local:
        logger.warning("Using void data upload instead of HTTP uploader")
        return VoidUploader(logger)
    logger.info("Using normal data uploader")
    return HTTPDataUploader(upload_url, logger)


class InputProvider:
    def get_input_type(self):
        """Return input type"""

    def get_mode(self):
        """Get mode"""
    
    def is_development_mode(self):
        """Is development mode"""

    def get_patient_id(self):
        pass

    def get_case_id(self):
        pass

    def get_reference(self):
        pass

    def skip_reports(self):
        pass

    def skip_download_upload(self):
        pass

    def get_input_resources_context(self):
        pass

    def __str__(self):
        return str(self.__dict__)

class StaticInputProvider(InputProvider):
    def __init__(self, mode, input_type, patient_id, case_id, 
            input_resources_context, 
            reference, skip_reports:bool=False, skip_download_upload:bool=False, 
            development_mode:bool=False) -> None:
        self.mode = mode
        self.input_type = input_type
        self.patient_id = patient_id
        self.case_id = case_id
        self.reference = reference
        self._skip_reports = skip_reports
        self._skip_download_upload = skip_download_upload
        self.development_mode = development_mode
        self.input_resources_context = input_resources_context

    def get_input_type(self):
        return self.input_type

    def get_mode(self):
        return self.mode

    def get_patient_id(self):
        return self.patient_id
    
    def get_case_id(self):
        return self.case_id
    
    def is_development_mode(self):
        return self.development_mode
    
    def get_reference(self):
        return self.reference
    
    def skip_reports(self):
        return self._skip_reports
    
    def skip_download_upload(self):
        return self._skip_download_upload
    
    def get_input_resources_context(self):
        return self.input_resources_context



import json
from eputils.thread.consumer import ConsumerThread
from ..utils.context import Context
from .message_processor import MessageProcessor
from .instruction import JobInstruction
from .job_provider import JobProvider

class RabbitMQJobProvider(JobProvider):
    def __init__(self, context: Context, message_processor: MessageProcessor):
        super().__init__()

        self._jobs = {}
        self._message_processor = message_processor
        self._logger = context.logger
        self._context = context

        self._fetcher = RabbitMQBasicConsumer(context.connection.host, 
                                              context.connection.creds_provider,
                                              context.messagebus.job_queue_name, 
                                              self._on_receive, 
                                              context).run()
    
    def get_job(self, id):
        return self._jobs[id]

    def _validate(self, data):
        if "message-type" not in data: 
            raise NotImplementedError("Cannot interpret message: No message type")
        if not data["message-type"] == "algorithm": 
            raise NotImplementedError("Cannot interpret message: Unknown message type")

    def _on_receive(self, data):
        json_data = json.loads(data)
        self._validate(json_data)

        input_resources: list = json_data["input-resources"]
        output_resources: list = json_data["output-resources"]
        sub_job_reference: str = json_data["sub-job-reference"]
        
        instruction = JobInstruction(sub_job_reference, input_resources, output_resources)

        self._message_processor.process(instruction)

class RabbitMQBasicConsumer:
    def __init__(self, host, creds_provider, queue_name, on_receive, context: Context):
        self._host = host
        self._creds_provider = creds_provider
        self._on_receive = on_receive
        self._queue_name = queue_name
        self._logger = context.logger
        self._context = context

    def run(self):
        '''
        Start polling action for fetching jobs
        '''
        self._thread = ConsumerThread(host=self._host, 
                                      queue_name=self._queue_name, 
                                      creds_provider=self._context.connection.creds_provider, 
                                      on_receive=self._on_receive, 
                                      app_id=f"{self._context.messagebus.app_id}.job-provider",
                                      logger=self._logger)
        self._context.add_thread(self._thread)
        
        self._logger.info("Starting RabbitMQ Consumer thread..")
        self._logger.info(f"Listening to queue {self._queue_name}..")
        self._thread.start()

        return self



import os
import shutil

def _fix_dirs(directory):
    if not os.path.exists(directory):
        os.makedirs(directory)

def file_exists(path):
    return os.path.exists(path)

def get_data_tree(dir):
    tree = dict()
    for f in os.listdir(dir):
        full_f = os.path.join(dir,f)
        if os.path.isdir(full_f):
            tree[f] = get_data_tree(full_f)
        else:
            tree[f] = 0
    
    return tree

def clear_folder(folder):
    if not os.path.exists(folder):
        print("Not a folder")
        return
    for filename in os.listdir(folder):
        file_path = os.path.join(folder, filename)
        try:
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)
        except Exception as e:
            print('Failed to delete %s. Reason: %s' % (file_path, e))

def get_file_contents(file, mode='r'):
    if not os.path.exists(file): return None
    with open(file, mode) as f:
        return f.read()

"""
Saving data to a file.

dir => should end with a `/`, i.e. `C:/some/path/` (and not `C:/some/path`)
filename => should only contain allowed filename tokens
"""
def write_to_file(dir, filename, data):
    _fix_dirs(dir)

    path = dir+filename

    with open(path, 'wb+') as f:
        f.write(data)


from datetime import datetime, timezone

def get_message_formatter():
    return StandardMessageFormatter()

class StandardMessageFormatter:
    def get_message(self, sub_job_reference, status, description) -> str:
        now = datetime.now(timezone.utc).isoformat()

        return f"""
        {{
            \"message-type\":\"algorithm-response\",
            \"date-time\":\"{now}\",
            \"sub-job-reference\":\"{sub_job_reference}\",
            \"status\":\"{status}\",
            \"description\":\"{description}\"
        }}
        """.strip()


from queue import SimpleQueue

from ..utils.manifest import Manifest
from ..utils.configuration import Configuration
from ..utils.message import get_message_formatter
from ..connection import get_messagebroker_host, get_creds_provider, get_algorithm_manager_host
from ..reporter.reporter_factory import get_preprocess_reporter
from ..reporter.reporter import Reporter

import os
import logging

host = get_messagebroker_host()
algorithm_manager_host = get_algorithm_manager_host()
creds_provider = get_creds_provider()
logging.basicConfig(level=os.environ.get("LOGLEVEL", "INFO"))
logger = logging.getLogger(__name__)
configuration = Configuration.get_configuration()
manifest = Manifest.get_manifest()
queue = SimpleQueue()
thread = None
message_formatter = get_message_formatter()
reporter = get_preprocess_reporter(algorithm_manager_host, logger, None)

def open_reporter() -> Reporter:
    """Get a reporter instance for reporting status updates to the Algorithm Manager"""
    return reporter


from .reporter import Reporter
from ..utils.http_tools import send_post_raw
from ..utils.message import get_message_formatter
from logging import Logger
import os

class RESTAPIReporter(Reporter):
    def __init__(self, 
            host: str,
            logger: Logger,
            route: str = "status-update") -> None:
        super().__init__(logger)
        self.host = host
        self.route = route
        self.report_url = f"http://{self.host}/{self.route}"
    
    def report(self, message) -> bool:
        url = self.report_url
        data = message
        self._print("Sending Rest POST Report")
        r = send_post_raw(url, data, self._logger, {"Content-Type": "application/json"})
        if r is not None:
            self._print("Sent Rest POST Report!")
            return True
        self._print("Failed Rest POST Report")
        return False

    def __enter__(self):
        return self
    
    def __exit__(self, type, value, tb):
        return super().__exit__(type, value, tb)

class PreprocessRESTAPIReporter(RESTAPIReporter):
    def __init__(self, 
            host: str,
            logger: Logger) -> None:
        super().__init__(host, logger)

        self.message_formatter = get_message_formatter()
    
    def report(self, message) -> bool:
        if "EP_ALGORITHM_SUB_JOB_REFERENCE" not in os.environ:
            raise Exception("Sub job reference not present in environment")

        sub_job_reference = os.environ["EP_ALGORITHM_SUB_JOB_REFERENCE"]

        processed_message = self.message_formatter.get_message(sub_job_reference, "busy", message)
        return super().report(processed_message)

    def __enter__(self):
        return super().__enter__()
    
    def __exit__(self, type, value, tb):
        return super().__exit__(type, value, tb)

class VoidReporter(Reporter):
    def report(self, message: str):
        return True
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, tb):
        return super().__exit__(type, value, tb)


import logging
from ..utils.http_tools import send_post_file
from ..utils.filesystem_utils import file_exists

class DataUploader:
    def send_request(self, patient_id, case_id, file_context, workdir, filename) -> bool:
        pass

class HTTPDataUploader(DataUploader):
    def __init__(self, upload_url: str, logger: logging.Logger) -> None:
        self.upload_url = upload_url
        self.logger = logger
    
    def send_request(self, patient_id, case_id, file_context, workdir, filename) -> bool:
        url = f"{self.upload_url}/upload/{patient_id}/{case_id}/{file_context}"
        resource_location = workdir+"/"+filename
        if not file_exists(resource_location): 
            self.logger.error(f"File {resource_location} does not exist")
            return False

        self.logger.info("Sending file post request..")
        r = send_post_file(url, resource_location, self.logger)
        self.logger.info("Posted file")

        return True

class VoidUploader(DataUploader):
    def __init__(self, logger: logging.Logger) -> None:
        self.logger = logger
    
    def send_request(self, patient_id, case_id, file_context, workdir, filename) -> bool:
        self.logger.info("Uploading file to void (so not uploading)")
        return True


from logging import Logger

class Reporter:
    def __init__(self, logger: Logger) -> None:
        self._logger = logger

    def _print(self, message):
        self._logger.info(message)
        
    def report(self, message: str) -> bool:
        raise NotImplementedError
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, tb):
        return

class ReporterConnectionDetails:
    def __init__(self,
            host,
            creds_provider = None,
            messagebus = None) -> None:
        self.host = host
        self.creds_provider = creds_provider
        self.messagebus = messagebus


from ..utils.manifest import Manifest
from ..reporter.reporter import Reporter
from ..reporter.rest_api_reporter import RESTAPIReporter
from ..connection import get_algorithm_manager_host
import logging


def __valid_version(version_int: int):
    return version_int == 1

def __get_reporter() -> Reporter:
    host = get_algorithm_manager_host()
    logger = logging.getLogger(__name__)
    route = "algorithm-validated"
    return RESTAPIReporter(host, logger, route)

def get_in_out_message(in_or_out: list):
    return ",".join([
        f"""{{
            \"data-name\": \"{x["data"]}\",
            \"store\": \"filesystem\",
            \"name\": \"{x["name"]}\"
        }}""" for x in in_or_out])

def validate():
    manifest = Manifest.get_manifest()

    if Manifest.Key.MANIFEST_VERSION not in manifest:
        raise KeyError(f"Manifest version not present! Does have: {dict(manifest)}")
    
    manifest_version: str = manifest[Manifest.Key.MANIFEST_VERSION]  # v1
    manifest_version_int = int(manifest_version.replace("v", ""))    # 1
    manifest_version_valid = __valid_version(manifest_version_int)
    if not manifest_version_valid:
        raise KeyError("Manifest version is not supported")

    all_keys = Manifest.Key.COMMAND in manifest \
                and Manifest.Key.IDENTIFIER in manifest \
                and Manifest.Key.INPUT in manifest \
                and Manifest.Key.OUTPUT in manifest
    if not all_keys: return False

    command: str = manifest[Manifest.Key.COMMAND]
    identifier: str = manifest[Manifest.Key.IDENTIFIER]
    input: list = manifest[Manifest.Key.INPUT]
    output: list = manifest[Manifest.Key.OUTPUT]

    if type(input) is not list:
        raise KeyError("Manifest input should be a list of strings")
    if type(output) is not list:
        raise KeyError("Manifest output should be a list of strings")
    if type(command) is not str:
        raise KeyError("Manifest command should be a string")
    if type(identifier) is not dict:
        raise KeyError("Manifest identifier should be a dict")

    input_message = get_in_out_message(input)
    output_message = get_in_out_message(output)

    identifier_name = identifier["name"]
    identifier_version = identifier["version"]

    # TODO: Register to API that validation was successful, including name of algorithm, version, all taken from manifest
    reporter = __get_reporter()
    with reporter:
        message = f"""
        {{
            \"message-type\":\"algorithm-validated\",
            \"report\": {{
                \"status\": \"valid\"
            }},
            \"algorithm-description\":{{
                \"identifier\":{{
                    \"name\": \"{identifier_name}\",
                    \"version\": \"{identifier_version}\"
                }},
                \"input-resources\": [
                    {input_message}
                ],
                \"output-resources\": [
                    {output_message}
                ]
            }}
        }}
        """.strip()

        print("Sending validation message..")

        reporter.report(message)

        print("Validation message sent!")


import os

EP_ENVIRONMENT_KEY = "EP_ENVIRONMENT"

DEFAULT_HOST = "localhost"
DOCKER_HOST = "rabbitmq"
DEFAULT_FILESYSTEM_HOST = "localhost:5101"
DOCKER_FILESYSTEM_HOST = "repository-filesystem:5101"
DEFAULT_ALGORITHMMANAGER_HOST = "localhost:5002"
DOCKER_ALGORITHMMANAGER_HOST = "algorithm-manager:5002"
DEFAULT_REGISTRY_HOST = "localhost:5080"
DOCKER_REGISTRY_HOST = "registry:5080"

def get_messagebroker_host():
    host = DEFAULT_HOST
    if EP_ENVIRONMENT_KEY not in os.environ:
        return host

    environment = os.environ[EP_ENVIRONMENT_KEY]
    if environment.find("_DOCKER") != -1:
        host = DOCKER_HOST
    
    return host

def get_filesystem_host():
    host = DEFAULT_FILESYSTEM_HOST
    if EP_ENVIRONMENT_KEY not in os.environ:
        return host

    environment = os.environ[EP_ENVIRONMENT_KEY]
    if environment.find("_DOCKER") != -1:
        host = DOCKER_FILESYSTEM_HOST
    
    return host

def get_algorithm_manager_host():
    host = DEFAULT_ALGORITHMMANAGER_HOST
    if EP_ENVIRONMENT_KEY not in os.environ:
        return host

    environment = os.environ[EP_ENVIRONMENT_KEY]
    if environment.find("_DOCKER") != -1:
        host = DOCKER_ALGORITHMMANAGER_HOST
    
    return host

def get_registry_host():
    host = DEFAULT_REGISTRY_HOST
    if EP_ENVIRONMENT_KEY not in os.environ:
        return host

    environment = os.environ[EP_ENVIRONMENT_KEY]
    if environment.find("_DOCKER") != -1:
        host = DOCKER_REGISTRY_HOST
    
    return host

def get_creds_provider():
    return lambda: ("admin", "admin")


from .reporter import Reporter, ReporterConnectionDetails
from ..utils.message import get_message_formatter
from eputils.thread.publisher import PublisherThread
from queue import SimpleQueue
from logging import Logger
import os

class AMTQReporter(Reporter):
    """Reports status updates to algorithm manager
    using AMTQ"""
    def __init__(self, logger: Logger, connection_details: ReporterConnectionDetails) -> None:
        super().__init__(logger)
        self.publish_queue = SimpleQueue()
        self.publisher_thread = PublisherThread(
            host=connection_details.host,
            creds_provider=connection_details.creds_provider,
            exchange=connection_details.messagebus.status_update_exchange,
            publish_queue=self.publish_queue,
            app_id=f"{connection_details.messagebus.app_id}.publisher",
            logger=logger)
        
    
    def report(self, message):
        self._print("Publishing "+ message)
        self.publish_queue.put((message, self._context.messagebus.status_update_routing))
        self._print("Published message")
        return True
    
    def __enter__(self):
        self.publisher_thread.start()
        return self
    
    def __exit__(self, type, value, tb):
        self.publisher_thread.stop()
        return super().__exit__(type, value, tb)

class PreprocessAMTQReporter(AMTQReporter):
    def __init__(self, logger: Logger, connection_details: ReporterConnectionDetails) -> None:
        super().__init__(logger, connection_details)
        self.message_formatter = get_message_formatter()

    def report(self, message):
        if self.publisher_thread is None:
            raise Exception("Not connected to RabbitMQ, call connect() first!")
        if "EP_ALGORITHM_SUB_JOB_REFERENCE" not in os.environ:
            raise Exception("Sub job reference not present in environment")
        if not self.publisher_thread.is_alive():
            raise Exception("Thread is not alive")
        if not self.publisher_thread.is_connected():
            raise Exception("Thread not connected to broker")

        sub_job_reference = os.environ["EP_ALGORITHM_SUB_JOB_REFERENCE"]

        processed_message = self.message_formatter.get_message(sub_job_reference, "busy", message)

        return super().report(processed_message)

    def __enter__(self):
        return super().__enter__()
    
    def __exit__(self, type, value, tb):
        return super().__exit__(type, value, tb)


from .rabbitmq_job_provider import RabbitMQJobProvider
from .static_job_provider import StaticJobProvider
from ..utils.input_provider import InputProvider

def get_job_provider(input_provider: InputProvider, context, message_processor):
    input_type = input_provider.get_input_type()
    if input_type == "environment" or input_type == "cli":
        return StaticJobProvider(
            context, 
            message_processor, 
            input_provider.get_reference(),
            input_provider.get_patient_id(),
            input_provider.get_case_id())
    elif input_type == "rabbitmq":
        return RabbitMQJobProvider(
            context, 
            message_processor)


from ..utils.context import Context
from ..utils.filesystem_utils import _fix_dirs

def get_resource_manager(context):
    return ResourceManager(context)

class ResourceManager:
    SUPPORTED_STORES = ["filesystem"]

    def __init__(self, context: Context):
        self._logger = context.logger
        self._repository_filesystem = RepositoryFileSystem(context)
        self._context = context

    def fetch_resource(self, resource_instruction, workdir):
        if resource_instruction is None:
            self._logger.warning("Fetching instruction was none!")
            return False

        store = resource_instruction["store"]
        self.__validate_store(store)
        
        if store == "filesystem":
            return self._repository_filesystem.fetch_resource(resource_instruction, workdir)
        return False
    
    def post_resource(self, resource_instruction, workdir):
        if resource_instruction is None:
            self._logger.warning("Posting instruction was none!")
            return False

        store = resource_instruction["store"]
        self.__validate_store(store)

        if store == "filesystem":
            return self._repository_filesystem.post_resource(resource_instruction, workdir)
    
    def __validate_store(self, store) -> bool:
        if store not in ResourceManager.SUPPORTED_STORES:
            raise NotImplementedError(f"Store {store} not supported")

    def _upload_resource_filesystem(self, workdir, filename):
        pass

class RepositoryFileSystem:
    def __init__(self, context: Context):
        self._logger = context.logger
        self.context = context

    def fetch_resource(self, resource_instruction, workdir):
        self._logger.info("Fetching data from filesystem..")

        data_provider = self.context.data_provider

        identification = resource_instruction["resource-identification"]
        patient_context = resource_instruction["patient-context"]
        patient_id = patient_context["patient-id"]
        case_id = patient_context["case-id"]
        filename = resource_instruction["resource-filename"]
        file_context = resource_instruction["resource-context"]

        request_body = {
            "patientId": patient_id,
            "caseId": case_id,
            "context": file_context,
            "fileName": filename
        }

        data = data_provider.send_request(request_body)
        if data is None: return False

        if data == True:
            self._logger.info("Data should be available")
            return True

        _fix_dirs(workdir)

        loc = f"{workdir}/{filename}"

        self._logger.info("Writing response..")
        with open(loc, 'wb+') as f:
            f.write(data)
        
        self._logger.info(f"Received {filename} from filesystem, saved into {loc}")
        return True
    
    def post_resource(self, resource_instruction, workdir):
        identification = resource_instruction["resource-identification"]
        patient_context = resource_instruction["patient-context"]
        patient_id = patient_context["patient-id"]
        case_id = patient_context["case-id"]
        filename = resource_instruction["resource-filename"]
        file_context = resource_instruction["resource-context"]

        data_uploader = self.context.data_uploader
        sent = data_uploader.send_request(
            patient_id=patient_id, 
            case_id=case_id, 
            file_context=file_context, 
            workdir=workdir, 
            filename=filename)

        return sent



import os

print("1")
# os.system("dir")
os.system("cd python-ep-algorithm-wrapper && py .\\.dev\\dist.py")

print("2")
os.system("cd ..")

os.system("py -m pip install --upgrade epalgorithmwrapper")
os.system("py -m pip install --upgrade epalgorithmwrapper") # Only picks last version after second call..?

print("3")
os.system("docker compose -f .\\local.docker-compose.build-template.yml build --no-cache")

print("Done!")



from logging import Logger
from threading import Thread
from .manifest import Manifest
from .input_provider import InputProvider
from ..dataprovider.dataprovider import DataProvider
from ..datauploader.datauploader import DataUploader

class ConnectionContext:
    def __init__(self,
                 host,
                 host_filesystem,
                 host_algorithm_manager,
                 creds_provider) -> None:
        self.host: str = host
        self.host_filesystem: str = host_filesystem
        self.host_algorithm_manager: str = host_algorithm_manager
        self.creds_provider = creds_provider

class MessagebusContext:
    def __init__(self,
                 app_id,
                 job_queue_name,
                 status_update_exchange,
                 status_update_routing) -> None:
        self.app_id = app_id
        self.job_queue_name = job_queue_name
        self.status_update_exchange = status_update_exchange
        self.status_update_routing = status_update_routing

class Context:
    def __init__(self, 
                 connection,
                 messagebus,
                 logger, 
                 manifest,
                 configuration,
                 input_provider,
                 data_provider,
                 data_uploader) -> None:
        self.input_provider: InputProvider = input_provider
        self.threads: list[Thread] = []
        self.connection: ConnectionContext = connection
        self.messagebus: MessagebusContext = messagebus
        self.logger: Logger = logger
        self.manifest: Manifest = manifest
        self.configuration = configuration
        self.data_provider: DataProvider = data_provider
        self.data_uploader: DataUploader = data_uploader

    def add_thread(self, thread: Thread):
        self.threads += [thread]

    def stop_threads(self):
        for t in self.threads:
            t.stop()


from enum import Enum
from .filesystem_utils import get_file_contents
import json

class Configuration(dict):
    CONFIGURATION_FILE_PATH = "./epalgorithmwrapper/configuration.json"

    class Key(Enum):
        INPUT_DIRECTORY="input-directory"
        OUTPUT_DIRECTORY="output-directory"
        OUTPUT_REPORT_LOCATION="output-report-location"
        REPORT_TEMPLATE_LOCATION="report-template-location"
        RABBITMQ_STATUS_UPDATE_EXCHANGE="rabbitmq-status-update-exchange"
        RABBITMQ_STATUS_UPDATE_ROUTING="rabbitmq-status-update-routing"

    
    def __init__(self, dict: dict) -> None:
        self.__dict__ = dict

    def __setitem__(self, key, item):
        self.__dict__[key.value] = item

    def __getitem__(self, key):
        return self.__dict__[key.value]

    def __repr__(self):
        return repr(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def __delitem__(self, key):
        del self.__dict__[key.value]

    def clear(self):
        return self.__dict__.clear()

    def copy(self):
        return self.__dict__.copy()

    def has_key(self, k):
        return k.value in self.__dict__

    def update(self, *args, **kwargs):
        return self.__dict__.update(*args, **kwargs)

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()   
             
    @staticmethod
    def get_configuration():
        configuration_str = get_file_contents(Configuration.CONFIGURATION_FILE_PATH)
        return Configuration(json.loads(configuration_str))